﻿unit fonctions_scaledpi;

interface

{$IFDEF FPC}
{$mode Delphi}
{$ENDIF}

uses
  Graphics, Forms,
  {$IFDEF VERSIONS}
    fonctions_version,
  {$ENDIF}
  Controls, Classes, sysutils;

Const
  FromDPI=8;//Screen.MenuFont.Size de la conception
  SPACING = 'Spacing';
  SCALE_NODE_HEIGHT = 'DefaultNodeHeight';
  SCALE_SEARCH_WIDTH = 'SearchWidth';
  SCALE_GLYPH_SIZE  = 'GlyphSize';
  {$IFDEF VERSIONS}
  gver_fonctions_scaledpi : T_Version = ( Component : 'Fonctions d''adaptation de fontes' ;
                                     FileUnit : 'fonctions_scaledpi' ;
              			                 Owner : 'Matthieu Giroux' ;
              			                 Comment : 'Adapt forms and controls to system.' ;
              			                 BugsStory :  'Version 1.0.2.0 : Changing Scale to ScaleValue, adding ScaleFont, global scale hidden.' + #13#10 +
                                                              'Version 1.0.1.2 : More components to scale.' + #13#10 +
                                                              'Version 1.0.1.1 : Never scale on design.' + #13#10 +
                                                              'Version 1.0.1.0 : Adding interfaces for classes.' + #13#10 +
                                                              'Version 1.0.0.0 : OK on linux with ini.' + #13#10 +
                                                              'Version 0.9.9.0 : OK on windows.' + #13#10 +
                                                              'Version 0.9.0.0 : To test.';
              			                 UnitType : 1 ;
              			                 Major : 1 ; Minor : 0 ; Release : 2 ; Build : 0 );

{$ENDIF}

type TFormScaleControl = procedure (const AControl : TComponent ; var Continue : Boolean ) of object;
     TFormAfterScaleControl = procedure (const AControl : TComponent ) of object;

procedure p_SetChildForm ( const afor_Reference: TCustomForm; const  afs_newFormStyle : TFormStyle ; const ab_Ajuster, ab_adjustconfirmed : boolean);

procedure p_addtoSoftware;
procedure HighDPI;
procedure ScaleDPIControl(const AForm, Control: TComponent); overload;
procedure ScaleDPIControl(const AForm, Control: TComponent;const ANewEchelle:Single); overload;
procedure ScaleFont(const AFont:TFont); overload;
procedure ScaleFont(const AFont:TFont;const ANewEchelle:Single); overload;
function ScaleValue(const Valeur:Integer;const ANewEchelle:Single):Integer;
procedure ScaleFormShow(const Control: TCustomForm;const ANewEchelle:Single);
procedure ScaleFormCreate(const Control: TCustomForm;const ANewEchelle:Single);
function fb_CalculateScale ( var AEchelle : Single ):Boolean;
procedure FormInScreen(const Control: TCustomForm);
procedure SetApplicationScale ( const asi_Scale : Single );
function  GetApplicationScale : Single ;

type
  INoAdaptComponent = interface
   ['{4D071431-71E6-4B55-9AD2-B815D34A8379}']
  End;

  ISpecialAdaptComponent = interface
    ['{4D071431-71E6-4B55-9AD2-B815D34A8379}']
    procedure ScaleComponent ( const NewScale : Single );
  End;

  { TDMAdaptForms }

  TDMAdaptForms = class(TDataModule)
    procedure ApplicationActivate ( Sender : TObject );
  private
    { private declarations }
  published
    { published declarations }
    constructor Create ( AOwner : TComponent ); override;

  end;

var
  gb_AdaptFormsToOS : Boolean = True;

implementation

uses fonctions_proprietes, math,
     typinfo, Grids,
     U_FormAdapt,
     fonctions_forms,
     CompSuperForm,
     fonctions_dialogs;

var ge_OldApplicationActivate : TNotifyEvent = nil;
    DMAdaptForms: TDMAdaptForms = nil;
    gsi_FontsScale:Single=1;


procedure SetApplicationScale ( const asi_Scale : Single );
Begin
  if asi_Scale <= 0 Then
   Begin
    MyShowMessage('You cannot Scale under 0.');
    Exit;
   end;
  gsi_FontsScale:=asi_Scale;
end;
function  GetApplicationScale : Single ;
Begin
  Result:=gsi_FontsScale;
end;

// Optional Application.OnActivate Event ( not using FormAdapt )
procedure TDMAdaptForms.ApplicationActivate ( Sender : TObject );
Begin
  if Assigned(ge_OldApplicationActivate) Then
    ge_OldApplicationActivate ( Sender );
  HighDPI;
end;

// no lfm ( not using FormAdapt )
constructor TDMAdaptForms.Create(AOwner: TComponent);
begin
  CreateNew(AOwner);
  if (ClassType <> TDataModule) and
     not (csDesigning in ComponentState) then
    begin
    if OldCreateOrder then
      DoCreate;
    end;
end;

// calculate ScaleValue from system ( windows only )
function fb_CalculateScale ( var AEchelle : Single ):Boolean;
var
  LNewEchelle : Single;
Begin
  if not gb_AdaptFormsToOS Then
  Begin
    Result := False;
    Exit;
  end;
  if Screen.MenuFont.Size = 0
    Then LNewEchelle:=FromDPI
    Else LNewEchelle:=Screen.MenuFont.Size;
  {WriteLn(IntToStr(Screen.SystemFont.Size));
  WriteLn(IntToStr(Screen.HintFont.Size));
  WriteLn(IntToStr(Screen.IconFont.Size));
  WriteLn(IntToStr(Screen.MenuFont.Size));}

  LNewEchelle:=LNewEchelle/FromDPI*AEchelle;
  Result := LNewEchelle<>AEchelle;
  AEchelle:=LNewEchelle;
End;

// Scale all Applications forms
procedure HighDPI;
var
  i: integer;
begin
  if fb_CalculateScale ( gsi_FontsScale ) then
    for i:=0 to Screen.FormCount-1 do
      ScaleFormShow(Screen.Forms[i],gsi_FontsScale);
end;

// Scale a value
function ScaleValue(const Valeur:Integer;const ANewEchelle:Single):Integer;
Var
  Ext:Single;
begin
  Ext:=Valeur*ANewEchelle;
  if Ext>=0 then
    Result:=Trunc(Ext+0.5)
  else
    Result:=Trunc(Ext-0.5);
end;
procedure ScaleFont(const AFont:TFont);
Begin
  ScaleFont(AFont,gsi_FontsScale);
end;

procedure ScaleFont(const AFont:TFont;const ANewEchelle:Single);
Begin
  with AFont do
   Begin
    if Size = 0 Then
     Size:=FromDPI;
    Size:=ScaleValue(Size,ANewEchelle);
   End;
end;

// Scale on Form Show
procedure ScaleFormShow(const Control: TCustomForm;const ANewEchelle:Single);
var
  i: integer;
Begin
  with Control do
  if not ( csDesigning in ComponentState ) //never change designing
   Then
    Begin
     {$IFDEF FPC}
     BeginUpdateBounds;
     {$ENDIF}

     ScaleFont(Font,ANewEchelle);
     for i:=0 to Control.ComponentCount-1 do
      ScaleDPIControl(Control,Control.Components[i],ANewEchelle);

     {$IFDEF FPC}
     EndUpdateBounds;
     {$ENDIF}
    End;
end;
// Scale on form create
procedure ScaleFormCreate(const Control: TCustomForm;const ANewEchelle:Single);
Begin
  with Control do
   Begin
     // resize if possible
     Width  := Min ( Screen.{$IFDEF WINDOWS}WorkAreaWidth{$ELSE}Width{$ENDIF}, ScaleValue(Width,ANewEchelle));
     // resize if possible
     Height  := Min ( Screen.{$IFDEF WINDOWS}WorkAreaHeight{$ELSE}Height{$ENDIF}, ScaleValue(Height,ANewEchelle));

   End;
  // form can be out of screen
  FormInScreen(Control);
end;

// form can be out of screen
procedure FormInScreen(const Control: TCustomForm);
Begin
  with Control do
   Begin
     // form can be out of screen
     if Width > Screen.{$IFDEF WINDOWS}WorkAreaWidth{$ELSE}Width{$ENDIF} Then
       Width   := Screen.{$IFDEF WINDOWS}WorkAreaWidth{$ELSE}Width{$ENDIF};
     if Height > Screen.{$IFDEF WINDOWS}WorkAreaHeight{$ELSE}Height{$ENDIF} Then
       Height  := Screen.{$IFDEF WINDOWS}WorkAreaHeight{$ELSE}Height{$ENDIF};
     if left + Width > Screen.{$IFDEF WINDOWS}WorkAreaWidth{$ELSE}Width{$ENDIF} Then
       Left   := Max ( 0, Screen.{$IFDEF WINDOWS}WorkAreaWidth{$ELSE}Width{$ENDIF} - Width );
     if Top + Height > Screen.{$IFDEF WINDOWS}WorkAreaHeight{$ELSE}Height{$ENDIF} Then
       Top    := Max ( 0, Screen.{$IFDEF WINDOWS}WorkAreaHeight{$ELSE}Height{$ENDIF} - Height );
   End;
end;

// Scale a font
procedure ScaleFontControl(const Control: TObject;const ANewEchelle:Single);
var  AFont : TFont;
Begin
  if ( Control is TCustomForm )
  or not fb_getComponentBoolProperty(Control,'ParentFont', False ) Then
   Begin
    AFont := TFont ( fobj_getComponentObjectProperty ( Control, 'Font' ));
    if assigned ( AFont ) then
     Begin
      ScaleFont(AFont,ANewEchelle);
     end;
   end;
end;

procedure p_ScaleChildControls(const AForm, Control: TComponent;const ANewEchelle:Single);
var WinControl: TWinControl;
    ChildControl : TControl;
    i: integer;
Begin
  if (Control is TWinControl) Then
  begin
    WinControl:=TWinControl(Control);
    for i:=0 to WinControl.ControlCount-1 do
     Begin
      ChildControl:= WinControl.Controls[i];
      if not ( ChildControl.Owner is TCustomForm ) then
       ScaleDPIControl(AForm, ChildControl,ANewEchelle);
     end;
  end;
end;

procedure ScaleDPIControl(const AForm, Control: TComponent);
Begin
  ScaleDPIControl(AForm, Control,gsi_FontsScale);
end;

// Scale a control
procedure ScaleDPIControl(const AForm, Control: TComponent;const ANewEchelle:Single);
var
  i: integer;
  AColumn : TCollection;
  AItem   : TCollectionItem;
  AContinue : Boolean;
begin
  AContinue := True;
  if assigned ( AForm )
  and (AForm is TF_FormAdapt ) Then
    (AForm as TF_FormAdapt ).p_AdaptControl ( Control, AContinue );
  if not AContinue Then
   Exit;
  if  {$IFNDEF FPC}Supports{$ENDIF}( Control {$IFDEF FPC}is{$ELSE},{$ENDIF} INoAdaptComponent)
  or ( csDesigning in Control.ComponentState )
  or ( ANewEchelle = 1 ) Then  //never change designing
   Exit;
  if Control is TControl Then
  with Control as TControl do
   begin
    with Constraints do
     begin
      MaxHeight:=ScaleValue(MaxHeight,ANewEchelle);
      MaxWidth:=ScaleValue(MaxWidth,ANewEchelle);
      MinHeight:=ScaleValue(MinHeight,ANewEchelle);
      MinWidth:=ScaleValue(MinWidth,ANewEchelle);
     end;

    if not ( Control is TCustomForm )
    and    ( Align   <> alClient    ) then
      begin
        if not (Align in [alTop,alBottom]) then
        begin
          if not (akRight in Anchors) and (Align<>alRight) then
            Left:=ScaleValue(Left,ANewEchelle);
          if ([akRight,akLeft]*Anchors<>[akRight,akLeft]) then
            Width:=ScaleValue(Width,ANewEchelle);
        end;
        if not (Align in [alLeft,alRight])then
        begin
          if (not (akBottom in Anchors))and(Align<>alBottom) then
            Top:=ScaleValue(Top,ANewEchelle);
          if [akBottom,akTop]*Anchors<>[akBottom,akTop] then
            Height:=ScaleValue(Height,ANewEchelle);
        end;
      end;

    if assigned ( GetPropInfo ( Control, SCALE_NODE_HEIGHT )) Then
      SetPropValue(Control, SCALE_NODE_HEIGHT, ScaleValue ( GetPropValue (Control, SCALE_NODE_HEIGHT ), ANewEchelle));
    if assigned ( GetPropInfo ( Control, SCALE_SEARCH_WIDTH )) Then
      SetPropValue(Control, SCALE_SEARCH_WIDTH, ScaleValue ( GetPropValue (Control, SCALE_SEARCH_WIDTH ), ANewEchelle));
    if assigned ( GetPropInfo ( Control, SPACING ))
    and ( PropType(Control,SPACING) in [tkInteger{$IFDEF FPC},tkQWord{$ENDIF},tkInt64] ) Then
     Begin
      I := ScaleValue ( GetPropValue (Control, SPACING ), ANewEchelle);
      if i < 0
       Then
        Begin
         I := Max ( -Width + 4, I );  // It is inside and text before picture spacing
         SetPropValue(Control, SPACING, i );
        end;
     end;
    if assigned ( GetPropInfo ( Control, SCALE_GLYPH_SIZE ))
    and ( PropType(Control,SCALE_GLYPH_SIZE) in [tkInteger{$IFDEF FPC},tkQWord{$ENDIF},tkInt64] ) Then
      SetPropValue(Control, SCALE_GLYPH_SIZE, ScaleValue ( GetPropValue (Control, SCALE_GLYPH_SIZE ), ANewEchelle));

    if ( Control is TCustomGrid )
    and assigned ( GetPropInfo ( Control, CST_PROPERTY_COLUMNS )) Then
     Begin
       AColumn := GetObjectProp ( Control, CST_PROPERTY_COLUMNS ) as TCollection;
       with AColumn do
       for i := 0 to Count - 1 do
        Begin
          AItem := Items [ i ];
          ScaleFontControl( AItem,ANewEchelle);
          if assigned ( GetPropInfo ( AItem, CST_PROPERTY_WIDTH )) Then
            SetPropValue ( AItem, CST_PROPERTY_WIDTH,
                           ScaleValue ( GetPropValue ( AItem, CST_PROPERTY_WIDTH )
                                  , ANewEchelle ));
        end;
     end;
  end;

  ScaleFontControl(Control,ANewEchelle);

  if {$IFNDEF FPC}Supports{$ENDIF}( Control {$IFDEF FPC}is{$ELSE},{$ENDIF} ISpecialAdaptComponent) Then
   ( Control as ISpecialAdaptComponent ).ScaleComponent(ANewEchelle);

  p_ScaleChildControls(AForm,Control,ANewEchelle);
  if assigned ( AForm )
  and (AForm is TF_FormAdapt ) Then
    (AForm as TF_FormAdapt ).p_AfterAdaptControl ( Control );

end;

// main optinal procedure ( not using FormAdapt )
procedure p_addtoSoftware;
Begin
  ge_OldApplicationActivate := Application.OnActivate;
  if DMAdaptForms = nil Then
    DMAdaptForms := TDMAdaptForms.Create(Application);
  Application.OnActivate    := TNotifyEvent ( fmet_getComponentMethodProperty ( DMAdaptForms, 'ApplicationActivate' ) );
end;


// procedure p_SetChildForm
// Setting FormStyle
// If using SuperForm, setting the child superform if not FormStyle is fsStayOnTop
// afor_Reference : Form Variable
// The style to set afs_newFormStyle
procedure p_SetChildForm(const afor_Reference: TCustomForm; const  afs_newFormStyle : TFormStyle; const ab_Ajuster, ab_adjustconfirmed : boolean );
var FBoxChilds:TWinControl;
const CST_BoxChilds = 'BoxChilds';
begin
  // Option ajuster
if ab_Ajuster
and ab_adjustconfirmed   then
  begin
  // Par dessus donc au centre
    if ( TFormStyle ( fli_getComponentProperty ( afor_Reference, 'FormStyle' )) = fsStayOnTop)
    and (    (TWindowState ( fli_getComponentProperty ( afor_Reference , 'WindowState' )) <> wsNormal         )
          or ( TPosition ( fli_getComponentProperty ( afor_Reference , 'Position' ))    <> poMainFormCenter )) then
      begin
        p_SetComponentProperty ( afor_Reference, 'Position', poMainFormCenter );
        p_SetComponentProperty ( afor_Reference, 'WindowState', wsNormal );
      end;

      // MDI enfant donc maximizée
    if  not gb_ModalStarted and ( afs_newFormStyle = fsMDIChild) then
      p_SetComponentProperty ( afor_Reference, 'WindowState', wsMaximized );
    // Mise à jour
    afor_Reference.Update ;
  end;

 if ( afor_Reference is TSuperForm )
  and not ( afs_newFormStyle in [fsStayOnTop]) Then
    with Application,MainForm do
    Begin
      BeginUpdateBounds;
    // Must use MainFormIni
    //       afor_Reference.AutoSize := True;
      ( afor_Reference as TSuperForm ).IncrustMode := aicAllClient;
      if   assigned ( GetPropInfo ( MainForm, CST_BoxChilds ))
      and  PropIsType      ( MainForm, CST_BoxChilds , tkClass) Then
        Begin
          FBoxChilds := TWinControl ( GetObjectProp   ( MainForm, CST_BoxChilds ));
          if not assigned ( FBoxChilds ) Then
            Begin
              FBoxChilds := TSuperBox.Create(MainForm);
              with FBoxChilds as TSuperBox do
                Begin
                  Parent := MainForm;
                  AutoScroll:=True;
                  Align:=alClient;
                end;
              SetObjectProp( MainForm, CST_BoxChilds, FBoxChilds );
            end;
    //       afor_Reference.Align := alClient;
         with FBoxChilds do
           Begin
             BeginUpdateBounds;
             with afor_Reference as TSuperForm do
               Begin
                ShowIncrust ( FBoxChilds );
               end;
             EndUpdateBounds;
           end;
        end
       Else
        ( afor_Reference as TSuperForm ).ShowIncrust ( MainForm );
       EndUpdateBounds;
     end
   else

  p_SetComponentProperty ( afor_Reference, 'FormStyle', afs_newFormStyle );
  {$IFDEF FPC}
  afor_Reference.Show;
  {$ENDIF}
  afor_Reference.BringToFront ;
end;   

initialization
{$IFDEF VERSIONS}
  p_ConcatVersion ( gver_fonctions_scaledpi );
{$ENDIF}
  ge_SetForm :={$IFNDEF FPC}@{$ENDIF}p_SetChildForm;
end.
